C**************************************************************************

      SUBROUTINE  dfs_print (root, lun)

C**************************************************************************
C
C      Subroutine DFS_PRINT is the iterative implementation of a
C      recursive, depth-first search algorithm that visits and
C      prints out the nodes in a call graph.
C
C
C      Pseudo-Code for the Recursive DFS Algorithm:
C
C          Initially, for all routines x:
C              visited[x] = false
C              locked[x] = false
C
C          PROC dfs (x, level)
C
C              LOCAL  y ;
C
C
C              IF locked[x] THEN
C                  Indent (level) ;
C                  Print "Routine <x> is recursively called." ;
C              ELSE
C                  IF visited[x] THEN
C                      Indent (level) ;  Print "<x> *" ;
C                      The subtree below x was expanded on the first
C                      visit - do not expand it on subsequent visits ;
C                  ELSE
C                      IF in-degree[x] > 1 THEN
C                          Indent (level) ;  Print "<x> *" ;
C                      ELSE
C                          Indent (level) ;  Print "<x>" ;
C                      ENDIF
C                      visited[x] = true ;
C                      locked[x] = true ;
C                      DO for each routine y that x calls
C                          CALL dfs (y, level+1) ;
C                      ENDDO
C                      locked[x] = false ;
C                  ENDIF
C              ENDIF
C
C          ENDPROC
C
C
C      Arguments:
C
C          ROOT  (Integer)
C              is the index in the routine table of the main module
C              at which the call tree traversal is to start.
C          LUN  (Integer)
C              is the FORTRAN logical unit number to which the output
C              structure chart is to be written.
C
C**************************************************************************


      IMPLICIT  NONE

C...      Parameters, global variables, and external definitions.

      INCLUDE  'struchart.inc'          ! STRUCHART definitions.
      INTEGER*4  trim                   ! External function.

C...      Subroutine arguments.

      INTEGER*4  root
      INTEGER*4  lun

C...      Local variables.

      INTEGER*4  length, level, ptr, routine





C...      Initialize the graph structure attributes.

      DO routine = 1, num_routines
          locked(routine) = .FALSE.
          visited(routine) = .FALSE.
      ENDDO


C...      Beginning at the main routine (the root of the tree),
C      traverse the graph, printing out the calling structure of
C      the program.

      length = trim (routine_name(root))
      WRITE (lun,500)  routine_name(root)(1:length)
500   FORMAT (//A, /<length>('-'), /)

      routine = root
      stack_pointer = 0


      DO WHILE (routine .GT. 0)


          length = trim (routine_name(routine))
          level = (stack_pointer/2) + 1
          ptr = NULL


C...      If the routine is locked (i.e., it also appears earlier in the
C      current search path), then a recursive call was made.  Note that
C      fact in the output and "return".

          IF (locked(routine)) THEN

              WRITE (lun,510)  routine_name(routine)(1:length) //
     +                         ' is recursively called.'
510           FORMAT (<(level-1)*8>X, A)

          ELSE

C...      If the routine has already been visited (i.e., it's a
C      utility-type function), then mark it in the output with an
C      asterisk and "return".  The routine's subtree was expanded
C      on the first visit.

              IF (visited(routine)) THEN

                  WRITE (lun,510)  routine_name(routine)(1:length) //
     +                             ' *'

              ELSE

C...      On the first visit to a routine, print out its name and perform
C      a depth-first search of its subtree.

                  IF (in_degree(routine) .GT. 1) THEN
                      WRITE (lun,510)  routine_name(routine)(1:length)
     +                                 // ' *'
                  ELSE
                      WRITE (lun,510)  routine_name(routine)(1:length)
                  ENDIF
                  visited(routine) = .TRUE.
                  ptr = calls(routine)

              ENDIF     ! Visited?

          ENDIF     ! Locked?


C...      If a routine has no subtree or the subtree does not need
C      to be expanded (both conditions indicated by PTR = NULL),
C      then "return".  If the expansion of the subtree of a routine
C      popped off the stack is complete, unlock the routine (i.e.,
C      it's no longer on the current search path) and "return" again.

          DO WHILE ((ptr .EQ. NULL) .AND. (stack_pointer .GT. 0))
              ptr = stack(stack_pointer)        ! Pop (Routine, Ptr)
              stack_pointer = stack_pointer - 1
              routine = stack(stack_pointer)
              stack_pointer = stack_pointer - 1
              ptr = adj_link(ptr)
              IF (ptr .EQ. NULL)  locked(routine) = .FALSE.
          ENDDO


C...      For each routine Y that X calls, lock X (i.e., X is on the
C      current search path), push X on the stack, and process Y.

          IF (ptr .EQ. NULL) THEN
              routine = -1
          ELSE
              locked(routine) = .TRUE.
              stack_pointer = stack_pointer + 1 ! Push (Routine, Ptr)
              stack(stack_pointer) = routine
              stack_pointer = stack_pointer + 1
              stack(stack_pointer) = ptr
              routine = adj_info(ptr)
          ENDIF

      ENDDO     ! Until stack empty and no node.


      RETURN


      END
