C**************************************************************************

      SUBROUTINE  GET_SYMBOLS (current_module, file_type,
     +                         rec, record_length)

C**************************************************************************
C
C      Subroutine GET_SYMBOLS extracts calling hierarchy information from
C      a VMS object record.  GET_SYMBOLS handles two types of VMS object
C      records: (i) main module header records and (ii) text information
C      and relocation records.  Other record types are ignored.
C
C      If a main module header (MHD) record is encountered, GET_SYMBOLS
C      extracts the module name, adds it to the routine table, and updates
C      the CURRENT_MODULE argument.
C
C      If a text information and relocation (TIR) record is encountered,
C      GET_SYMBOLS scans the Linker instructions in the record, looking
C      for references to global symbols.  For each reference found,
C      GET_SYMBOLS adds a "<current module> calls <global symbol>"
C      relation to the call graph.
C
C      NOTE:  GET_SYMBOLS is dependent upon the VMS object file format,
C      which is described in the VAX/VMS Linker Manual.  Changes in
C      the file format might require changes in GET_SYMBOLS.
C
C
C      Example Usage (in Pseudo-Code):
C
C          Open the object file
C          current_module = ' '
C          DO WHILE (NOT EOF)
C              Read an object record from the object file
C              CALL get_symbols (current_module, 'OBJ',
C                                object_record, length)
C          ENDDO
C
C
C      Arguments:
C
C          CURRENT_MODULE  (Character)
C              is the name of the current module.  Initially this should
C              be spaced out (' ').  As GET_SYMBOLS processes main module
C              header records, it will pull out the module name and store
C              it in this argument.
C          FILE_TYPE  (Character)
C              is the type of the file from which the object record
C              was read: 'OBJ' for a ".OBJ" object file, and 'OLB' for
C              a ".OLB" object library.  (See the DEFINED_IN field of
C              the routine table in the STRUCHART.INC include file.)
C          REC  (Object record structure)
C              is the object record GET_SYMBOLS will scan.
C          RECORD_LENGTH  (Integer)
C              is the length of the object record.
C
C**************************************************************************


      IMPLICIT  NONE

C...      Parameters, global variables, and external definitions.

      INCLUDE '($MHDEF)'        ! Module header record definition.
      INCLUDE '($OBJRECDEF)'    ! Object language record definition.
      INCLUDE '($TIRDEF)'       ! Text information and relocation record.
      INCLUDE  'struchart.inc'  ! STRUCHART definitions.
      INTEGER*4  add_routine    ! External function.

C...      Subroutine arguments.

      CHARACTER*(*)  current_module
      CHARACTER*(*)  file_type  ! 'OBJ' or 'OLB'.

      STRUCTURE  /OBJECT_RECORD/
          UNION
              MAP
                  CHARACTER  text*(OBJ$C_MAXRECSIZ)
              END MAP
              MAP
                  BYTE       buffer(OBJ$C_MAXRECSIZ)
              END MAP
              MAP
                  RECORD  /MHDEF/  mhd          ! Module header record.
              END MAP
              MAP
                  RECORD  /OBJRECDEF/  obj      ! Object file record.
              END MAP
              MAP
                  RECORD  /TIRDEF/  tir         ! Text info/reloc record.
              END MAP
          END UNION
      END STRUCTURE

      RECORD  /OBJECT_RECORD/  rec

      INTEGER*4  record_length

C...      Local variables.

      CHARACTER*32  symbol
      INTEGER*4  calling_routine, command, i, length, ptr

C**************************************************************************
C
C      Data Length Table:
C
C          The Data Length Table defines the lengths of the data fields
C      for the various Text Information and Relocation commands recognized
C      by the VMS Linker.  Subroutine GET_SYMBOLS uses these lengths to
C      step through the command stream in a TIR record.  Certain types of
C      commands which have variable length data fields are flagged with
C      a negative length in the table:
C
C            -1 : Standard name (first byte = # of characters that follow)
C            -2 : Standard name + byte + argument descriptor
C            -3 : Word + standard name
C            -4 : Like standard name (first byte = # of bytes that follow)
C
C**************************************************************************

      INTEGER*4  data_length(0:127)

      DATA  data_length(TIR$C_STA_GBL)          /-1/
      DATA  data_length(TIR$C_STA_SB)           / 1/
      DATA  data_length(TIR$C_STA_SW)           / 2/
      DATA  data_length(TIR$C_STA_LW)           / 4/
      DATA  data_length(TIR$C_STA_PB)           / 2/
      DATA  data_length(TIR$C_STA_PW)           / 3/
      DATA  data_length(TIR$C_STA_PL)           / 5/
      DATA  data_length(TIR$C_STA_UB)           / 1/
      DATA  data_length(TIR$C_STA_UW)           / 2/
      DATA  data_length(TIR$C_STA_BFI)          / 0/
      DATA  data_length(TIR$C_STA_WFI)          / 0/
      DATA  data_length(TIR$C_STA_LFI)          / 0/
                        ! Entry point mask from GSD$C_EPM subrecord?
      DATA  data_length(TIR$C_STA_EPM)          /-1/
      DATA  data_length(TIR$C_STA_CKARG)        /-2/
      DATA  data_length(TIR$C_STA_WPB)          / 3/
      DATA  data_length(TIR$C_STA_WPW)          / 4/
      DATA  data_length(TIR$C_STA_WPL)          / 6/
      DATA  data_length(TIR$C_STA_LSY)          /-3/
      DATA  data_length(TIR$C_STA_LIT)          / 1/
                        ! Entry point mask from GSD$C_LEPM record?
      DATA  data_length(TIR$C_STA_LEPM)         /-3/
      DATA  data_length(TIR$C_STO_SB)           / 0/
      DATA  data_length(TIR$C_STO_SW)           / 0/
      DATA  data_length(TIR$C_STO_L)            / 0/    ! TIR$C_STO_LW
      DATA  data_length(TIR$C_STO_BD)           / 0/
      DATA  data_length(TIR$C_STO_WD)           / 0/
      DATA  data_length(TIR$C_STO_LD)           / 0/
      DATA  data_length(TIR$C_STO_LI)           / 0/
      DATA  data_length(TIR$C_STO_PIDR)         / 0/
      DATA  data_length(TIR$C_STO_PICR)         / 0/
      DATA  data_length(TIR$C_STO_RSB)          / 0/
      DATA  data_length(TIR$C_STO_RSW)          / 0/
      DATA  data_length(TIR$C_STO_RL)           / 0/
      DATA  data_length(TIR$C_STO_VPS)          / 2/
      DATA  data_length(TIR$C_STO_USB)          / 0/
      DATA  data_length(TIR$C_STO_USW)          / 0/
      DATA  data_length(TIR$C_STO_RUB)          / 0/
      DATA  data_length(TIR$C_STO_RUW)          / 0/
      DATA  data_length(TIR$C_STO_B)            / 0/
      DATA  data_length(TIR$C_STO_W)            / 0/
      DATA  data_length(TIR$C_STO_RB)           / 0/
      DATA  data_length(TIR$C_STO_RW)           / 0/
      DATA  data_length(TIR$C_STO_RIVB)         /-4/
      DATA  data_length(TIR$C_STO_PIRR)         / 0/
      DATA  data_length(TIR$C_OPR_NOP)          / 0/
      DATA  data_length(TIR$C_OPR_ADD)          / 0/
      DATA  data_length(TIR$C_OPR_SUB)          / 0/
      DATA  data_length(TIR$C_OPR_MUL)          / 0/
      DATA  data_length(TIR$C_OPR_DIV)          / 0/
      DATA  data_length(TIR$C_OPR_AND)          / 0/
      DATA  data_length(TIR$C_OPR_IOR)          / 0/
      DATA  data_length(TIR$C_OPR_EOR)          / 0/
      DATA  data_length(TIR$C_OPR_NEG)          / 0/
      DATA  data_length(TIR$C_OPR_COM)          / 0/
      DATA  data_length(TIR$C_OPR_INSV)         / 2/
      DATA  data_length(TIR$C_OPR_ASH)          / 0/
      DATA  data_length(TIR$C_OPR_USH)          / 0/
      DATA  data_length(TIR$C_OPR_ROT)          / 0/
      DATA  data_length(TIR$C_OPR_SEL)          / 0/
      DATA  data_length(TIR$C_OPR_REDEF)        /-1/
      DATA  data_length(TIR$C_OPR_DFLIT)        / 1/
      DATA  data_length(TIR$C_CTL_SETRB)        / 0/
      DATA  data_length(TIR$C_CTL_AUGRB)        / 4/
      DATA  data_length(TIR$C_CTL_DFLOC)        / 0/
      DATA  data_length(TIR$C_CTL_STLOC)        / 0/
      DATA  data_length(TIR$C_CTL_STKDL)        / 0/


C...      Lookup the current module name in the routine table and
C      save its index.

      IF (current_module .NE. ' ') THEN
          calling_routine = add_routine (current_module, ' ')
      ENDIF


C...      Main module header record.  Extract the module name and
C      enter it in the routine table.

      IF ((rec.obj.OBJ$B_RECTYP .EQ. OBJ$C_HDR) .AND.
     +    (rec.obj.OBJ$B_SUBTYP .EQ. MHD$C_MHD)) THEN

          length = ZEXT (rec.mhd.MHD$B_NAMLNG)
          current_module = rec.mhd.MHD$T_NAME(1:length)
          calling_routine = add_routine (current_module, file_type)


C...      Text information and relocation record.  Extract the global
C      symbols referenced in the record.

      ELSEIF (rec.obj.OBJ$B_RECTYP .EQ. OBJ$C_TIR) THEN

          i = 2

          DO WHILE (i .LE. record_length)

              command = rec.buffer(I)
              i = i + 1         ! Skip command byte.

C...      Extract the global symbol and add the module-calls-symbol
C      relation to the call graph.

              IF (command .EQ. TIR$C_STA_GBL) THEN

                  length = ZEXT (rec.buffer(i))
                  symbol = rec.text(i+1:i+length)

                  IF (exclude_flag .EQ. P_DEFAULTS) THEN
                      IF (symbol(1:4) .EQ. 'FOR$')  symbol = ' '
                      IF (symbol(1:4) .EQ. 'MTH$')  symbol = ' '
                      IF (symbol(1:4) .EQ. 'OTS$')  symbol = ' '
                      IF (symbol(1:4) .EQ. 'STR$')  symbol = ' '
                  ELSEIF (exclude_flag .EQ. P_DOLLARS) THEN
                      IF (INDEX (symbol, '$') .GT. 0)  symbol = ' '
                  ENDIF

                  IF (symbol .NE. ' ') THEN             ! Not excluded?
                      ptr = add_routine (symbol, ' ')
                      CALL add_node (calling_routine, ptr)
                  ENDIF

              ENDIF

C...      Step over the TIR command's operands to the next command.

              IF (command .LT. 0) THEN  ! Store immediate command?
                  i = i + ABS (command)
              ELSE
                  length = data_length(command)
                  IF (length .EQ. -1) THEN
                      length = ZEXT (rec.buffer(i))
                      i = i + 1 + length
                  ELSEIF (length .EQ. -2) THEN
                      length = ZEXT (rec.buffer(i))
                      i = i + 1 + length + 1 + 1
                      length = ZEXT (rec.buffer(i))
                      i = i + 1 + length
                  ELSEIF (length .EQ. -3) THEN
                      i = i + 2         ! Skip word.
                      length = ZEXT (rec.buffer(i))
                      i = i + 1 + length
                  ELSEIF (length .EQ. -4) THEN
                      length = ZEXT (rec.buffer(i))
                      i = i + 1 + length
                  ELSE
                      i = i + length
                  ENDIF
              ENDIF     ! If not a Store Immediate command.

          ENDDO     ! Until the contents of the record are exhausted.

      ENDIF     ! If a TIR record.


      RETURN


      END
