C*******************************************************************************

      SUBROUTINE  getword (string, start, delimiters,
     +                     first, last, length)

C*******************************************************************************
C
C      Subroutine GETWORD gets the next "word" from a string of text.
C      A "word"  is a string of characters delimited on each side by any
C      character which is in a set of specified delimiters or by the end
C      of the string.
C
C      A "word" has the following syntax (< > means optional):
C
C            <delimiters>  non-delimiters  delimiters|end-of-string
C
C      GETWORD uses Runtime Library routines STR$FIND_FIRST_NOT_IN_SET
C      and STR$FIND_FIRST_IN_SET to:
C
C            1) starting at STRING(START:), span delimiters,
C               resulting in the first character of the next "word"
C            2) starting at STRING(FIRST:), span non-delimiters,
C               resulting in the last character of the next "word".
C
C      GETWORD returns the length and location of the next "word" to the
C      calling routine.  If a "word" is not found or an error condition
C      is detected, GETWORD returns a null string (length = 0) located
C      one character position past the end of the input string.
C
C
C      Example Usage:
C
C          delimiters = ' ,'    ! Words separated by blanks or commas.
C          start = 1
C          CALL getword (string, start, delimiters, first, last, length)
C          DO WHILE (length .GT. 0)
C              ... process STRING(FIRST:LAST) ...
C              start = last + 1
C              CALL getword (string, start, delimiters, first, last, length)
C          ENDDO
C
C
C      Arguments:
C
C          STRING  (Character)
C              The character string to be scanned.
C          START  (Integer)
C              The position (1..n) within the string from which the
C              scan is to start.  This argument is checked for
C              validity (including an implied check that the length
C              of the input string is greater than zero).
C          DELIMITERS  (Character)
C              A character string, each character of which is to be
C              considered a word delimiter; the beginning and the
C              end of the input string are implied delimiters.
C          FIRST  (Integer)
C              Returns the position (1..n) within the string at
C              which the next "word" begins; equals the length of
C              STRING plus one if no next "word" is found.
C          LAST  (Integer)
C              Returns the position (1..n) within the string at
C              which the next "word" ends; equals FIRST - 1 if no
C              next "word" is found.
C          LENGTH  (Integer)
C              Returns the length in characters of the next "word";
C              equals zero if no next "word" is found or START is
C              invalid.
C
C*******************************************************************************


      IMPLICIT NONE

C...      Parameters and external definitions.

      INTEGER*4  STR$FIND_FIRST_IN_SET          ! External functions.
      INTEGER*4  STR$FIND_FIRST_NOT_IN_SET

C...      Subroutine arguments.

      CHARACTER*(*)  string, delimiters
      INTEGER*4  start, first, last, length

C...      Local variables.

      INTEGER*4  i





C...      Get the length of the input string and set up the first and
C      last pointers as if the next word were NOT found.

      length = LEN (string)             ! Length of input string.
      first = length + 1
      last = first - 1

C...      If the start position is invalid, then return "Not Found".

      IF ((start .LT. 1) .OR. (length .LT. start)) THEN
          length = 0
          RETURN
      ENDIF

C...      Skip delimiters to reach the start of the next "word".

      i = STR$FIND_FIRST_NOT_IN_SET (string(start:), delimiters)
      IF (i .GT. 0) THEN
          first = start + i - 1         ! Start of word found.
      ELSE
          length = 0                    ! End of string.
          RETURN
      ENDIF

C...      Skip non-delimiters to reach the end of the "word".

      i = STR$FIND_FIRST_IN_SET (string(first:), delimiters)
      IF (i .GT. 0) THEN
          last = first + i - 2          ! End of word found.
      ELSE
          last = LEN (string)           ! End of string.
      ENDIF


      RETURN


      END
