C**************************************************************************

      PROGRAM  STRUCHART

C**************************************************************************
C
C      STRUCHART is an automatic structure chart generator.  It reads the
C      VAX/VMS object modules making up a program (or multiple programs),
C      extracts the global symbols defined in each object module, and builds
C      the structure chart(s) defined by that symbol information.
C
C      To compile and link STRUCHART, run the following command procedure:
C
C          $ @BUILDSC.COM
C
C      Before running STRUCHART, first add it to your process command
C      table using the VMS Command Definition Utility:
C
C          $ SET COMMAND STRUCHART
C
C      Then, to invoke STRUCHART, enter:
C
C          STRUCHART [/[NO]DOLLARS] [/MAIN=(<name>[,...])]
C                    [/OUTPUT[=<filename>]]  [<wildcard_spec> [,...]]
C
C      where
C
C          [NO]DOLLARS - controls the inclusion of module names with
C              embedded dollar signs.  NODOLLARS excludes ALL module
C              names with embedded dollar signs; DOLLARS includes ALL
C              module names with embedded dollar signs.  If neither is
C              specified, STRUCHART excludes names according to its
C              default list of module name prefixes.
C
C          MAIN=(<name>,...) - specifies one or more main module names.
C              If no main module names are specified, STRUCHART will
C              print out a structure chart for each module which is
C              not called by any other.  Otherwise, STRUCHART will only
C              print out structure charts whose roots are the specified
C              module names.
C
C          OUTPUT=<filename> - specifies a file to receive the structure
C              chart output.  If this option is not specified, the
C              output is written to SYS$OUTPUT.  If "/OUTPUT" is
C              specified without a file name, the file name defaults
C              to "<main>.SCO", where <main> is the name of the main
C              module in the structure chart.
C
C          <wildcard_spec>, ...
C              specify STRUCHART's input files.  Each argument can be
C              a single file or a wildcard file specification.  Object
C              libraries with a file extension of ".OLB" are scanned
C              and treated as if a set of separate object files were
C              specified.  Intermediate files with a file extension of
C              ".SCI" are read as text files containing module relations.
C              All other input files have their file extension replaced
C              by ".OBJ"; e.g., you can say "*.FOR" and STRUCHART will
C              pick up the corresponding ".OBJ" files.
C
C**************************************************************************


      IMPLICIT  NONE

C...      Parameters, global variables, and external definitions.

      INCLUDE  'struchart.inc'  ! STRUCHART definitions.
      PARAMETER  P_SCO_LUN = 9  ! Logical unit number for output file.
      EXTERNAL  CLI$_ABSENT, CLI$_NEGATED, CLI$_PRESENT
      INTEGER*4  fparse, trim   ! External functions.
      INTEGER*4  CLI$GET_VALUE, CLI$PRESENT, LIB$FIND_FILE

C...      Local variables.

      CHARACTER  extension*4, main*32, message*128
      CHARACTER  output_file*128, source_file*128
      CHARACTER  sticky_defaults*128, wildcard_spec*128
      INTEGER*4  context, i, length, status
      LOGICAL  name_from_module


C...      Set flags based on command line options.

      status = CLI$PRESENT ('DOLLARS')
      IF (status .EQ. %LOC(CLI$_PRESENT)) THEN
          exclude_flag = P_NONE         ! /DOLLARS - includes all $-routines.
      ELSEIF (status .EQ. %LOC(CLI$_NEGATED)) THEN
          exclude_flag = P_DOLLARS      ! /NODOLLARS - excludes all $-routines.
      ELSE
          exclude_flag = P_DEFAULTS     ! Otherwise, exclude default list only.
      ENDIF


C...      Initialize the adjacency list and the routine table to empty.

      NUM_ADJ = 0
      NUM_ROUTINES = 0


C...      Get the input file specifications from the command line.
C      Read each file and build the caller-callee data structure.

      sticky_defaults = ' '

      DO WHILE (CLI$GET_VALUE ('P1', wildcard_spec) .NE.
     +          %LOC(CLI$_ABSENT))

          IF (fparse (wildcard_spec, %DESCR(';'), sticky_defaults, , ,
     +                source_file)) THEN
              wildcard_spec = source_file
          ENDIF

          context = 0

          DO WHILE (LIB$FIND_FILE (wildcard_spec, source_file, context))

              WRITE (6,*)  source_file(1:trim(source_file))

              CALL fparse (source_file, , , %DESCR('TYPE'), , extension)

              IF (extension .EQ. '.OLB') THEN
                  CALL scan_olb (source_file)           ! Object library.
              ELSEIF (extension .EQ. '.SCI') THEN
                  CALL read_sci (source_file)           ! Intermediate file.
              ELSE
                  CALL read_obj (source_file)           ! Object file.
              ENDIF

          ENDDO

          sticky_defaults = wildcard_spec

      ENDDO


C...      Open up the output file.  If "/OUTPUT" was specified, base
C    the output file name on the main module name.  If "/OUTPUT=<filename>"
C    was specified, open <filename> for output.  If neither was specified,
C    write the structure charts out to SYS$OUTPUT.

      status = CLI$PRESENT ('OUTPUT')
      IF (status .EQ. %LOC(CLI$_PRESENT)) THEN
          status = CLI$GET_VALUE ('OUTPUT', output_file)
          IF (status .EQ. %LOC(CLI$_ABSENT)) THEN
              name_from_module = .TRUE.         ! "/OUTPUT"
          ELSE
              name_from_module = .FALSE.        ! "/OUTPUT=<filename>"
              OPEN (UNIT=P_SCO_LUN, NAME=output_file, TYPE='NEW',
     +              CARRIAGECONTROL='LIST', IOSTAT=status, ERR=90)
          ENDIF
      ELSE
          name_from_module = .FALSE.            ! Option not specified.
          output_file = 'SYS$OUTPUT'
          OPEN (UNIT=P_SCO_LUN, NAME=output_file, TYPE='NEW',
     +          CARRIAGECONTROL='LIST', IOSTAT=status, ERR=90)
      ENDIF


C...      If the names of one or more main modules was specified on the
C      command line, generate a structure chart for each of the programs.
C      For each of the main routines, traverse that routine's subtree and
C      output its structure chart.  If the output file names are to be
C      based on the main module names, open a new output file for each
C      main routine.

      IF (CLI$PRESENT ('MAIN') .NE. %LOC(CLI$_ABSENT)) THEN

          DO WHILE (CLI$GET_VALUE ('MAIN', main) .NE. %LOC(CLI$_ABSENT))
              DO i = 1, num_routines
                  IF (routine_name(i) .EQ. main) THEN
                      IF (name_from_module) THEN
                          CALL fparse (routine_name(i), %DESCR('.SCO'),
     +                                  , , , output_file)
                          OPEN (UNIT=P_SCO_LUN, NAME=output_file,
     +                          TYPE='NEW', CARRIAGECONTROL='LIST',
     +                          IOSTAT=status, ERR=90)
                      ENDIF
                      CALL dfs_print (i, P_SCO_LUN)
                      IF (name_from_module)  CLOSE (UNIT=P_SCO_LUN)
                  ENDIF
              ENDDO
          ENDDO

      ELSE

C...      If no main modules were specified on the command line, generate
C      structure charts for all possible programs.  For each "root" (a node
C      called by no other node) in the graph, traverse the tree below the
C      root and display that root's structure chart.  If the output file
C      names are to be based on the main module names, open a new output
C      file for each root.

          DO i = 1, num_routines
              IF ((called_by(i) .EQ. NULL) .AND.        ! A root node?
     +            (defined_in(i) .NE. 'OLB')) THEN      ! Not a library routine?
                  IF (name_from_module) THEN
                      CALL fparse (routine_name(i), %DESCR('.SCO'),
     +                              , , , output_file)
                      OPEN (UNIT=P_SCO_LUN, NAME=output_file,
     +                      TYPE='NEW', CARRIAGECONTROL='LIST',
     +                      IOSTAT=status, ERR=90)
                  ENDIF
                  CALL dfs_print (i, P_SCO_LUN)
                  IF (name_from_module)  CLOSE (UNIT=P_SCO_LUN)
              ENDIF
          ENDDO

      ENDIF


C...      If there was an error opening up the output file, display
C      an error message.

      status = 0                ! No errors.

90    IF (status .NE. 0) THEN
          CALL ERRSNS ( , , , , status)
          WRITE (6,*)  'Error opening output file: "',
     +                 output_file(1:trim(output_file)), '"'
          CALL LIB$SYS_GETMSG (status, length, message, , )
          WRITE (6,*)  message(1:length)
          CALL EXIT
      ENDIF


      END
